package java.time.chrono;

import java.util.Collections;
import java.util.Arrays;
import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;
import java.io.ObjectStreamException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.time.temporal.TemporalQueries;
import java.time.temporal.TemporalAccessor;
import java.time.temporal.Temporal;
import java.time.temporal.ValueRange;
import java.time.temporal.ChronoField;
import java.time.DateTimeException;
import java.time.temporal.TemporalAmount;
import java.time.temporal.UnsupportedTemporalTypeException;
import java.time.temporal.ChronoUnit;
import java.util.Objects;
import java.time.temporal.TemporalUnit;
import java.util.List;
import java.io.Serializable;

final class ChronoPeriodImpl implements ChronoPeriod, Serializable
{
    private static final long serialVersionUID = 57387258289L;
    private static final List<TemporalUnit> SUPPORTED_UNITS;
    private final Chronology chrono;
    final int years;
    final int months;
    final int days;
    
    ChronoPeriodImpl(Chronology chronology, int years, int months, int days) {
        Objects.requireNonNull(chronology, "chrono");
        this.chrono = chronology;
        this.years = years;
        this.months = months;
        this.days = days;
    }
    
    @Override
    public long get(TemporalUnit obj) {
        if (obj == ChronoUnit.YEARS) {
            return (long)this.years;
        }
        if (obj == ChronoUnit.MONTHS) {
            return (long)this.months;
        }
        if (obj == ChronoUnit.DAYS) {
            return (long)this.days;
        }
        throw new UnsupportedTemporalTypeException("Unsupported unit: " + obj);
    }
    
    @Override
    public List<TemporalUnit> getUnits() {
        return ChronoPeriodImpl.SUPPORTED_UNITS;
    }
    
    @Override
    public Chronology getChronology() {
        return this.chrono;
    }
    
    @Override
    public boolean isZero() {
        return this.years == 0 && this.months == 0 && this.days == 0;
    }
    
    @Override
    public boolean isNegative() {
        return this.years < 0 || this.months < 0 || this.days < 0;
    }
    
    @Override
    public ChronoPeriod plus(TemporalAmount temporalAmount) {
        ChronoPeriodImpl validateAmount = this.validateAmount(temporalAmount);
        return new ChronoPeriodImpl(this.chrono, Math.addExact(this.years, validateAmount.years), Math.addExact(this.months, validateAmount.months), Math.addExact(this.days, validateAmount.days));
    }
    
    @Override
    public ChronoPeriod minus(TemporalAmount temporalAmount) {
        ChronoPeriodImpl validateAmount = this.validateAmount(temporalAmount);
        return new ChronoPeriodImpl(this.chrono, Math.subtractExact(this.years, validateAmount.years), Math.subtractExact(this.months, validateAmount.months), Math.subtractExact(this.days, validateAmount.days));
    }
    
    private ChronoPeriodImpl validateAmount(TemporalAmount obj) {
        Objects.requireNonNull(obj, "amount");
        if (!(obj instanceof ChronoPeriodImpl)) {
            throw new DateTimeException("Unable to obtain ChronoPeriod from TemporalAmount: " + obj.getClass());
        }
        ChronoPeriodImpl chronoPeriodImpl = (ChronoPeriodImpl)obj;
        if (!this.chrono.equals(chronoPeriodImpl.getChronology())) {
            throw new ClassCastException("Chronology mismatch, expected: " + this.chrono.getId() + ", actual: " + chronoPeriodImpl.getChronology().getId());
        }
        return chronoPeriodImpl;
    }
    
    @Override
    public ChronoPeriod multipliedBy(int y) {
        if (this.isZero() || y == 1) {
            return this;
        }
        return new ChronoPeriodImpl(this.chrono, Math.multiplyExact(this.years, y), Math.multiplyExact(this.months, y), Math.multiplyExact(this.days, y));
    }
    
    @Override
    public ChronoPeriod normalized() {
        long monthRange = this.monthRange();
        if (monthRange > 0L) {
            long n = (long)this.years * monthRange + (long)this.months;
            long value = n / monthRange;
            int n2 = (int)(n % monthRange);
            if (value == (long)this.years && n2 == this.months) {
                return this;
            }
            return new ChronoPeriodImpl(this.chrono, Math.toIntExact(value), n2, this.days);
        }
        else {
            return this;
        }
    }
    
    private long monthRange() {
        ValueRange range = this.chrono.range(ChronoField.MONTH_OF_YEAR);
        if (range.isFixed() && range.isIntValue()) {
            return range.getMaximum() - range.getMinimum() + 1L;
        }
        return -1L;
    }
    
    @Override
    public Temporal addTo(Temporal temporal) {
        this.validateChrono(temporal);
        if (this.months == 0) {
            if (this.years != 0) {
                temporal = temporal.plus((long)this.years, ChronoUnit.YEARS);
            }
        }
        else {
            long monthRange = this.monthRange();
            if (monthRange > 0L) {
                temporal = temporal.plus((long)this.years * monthRange + (long)this.months, ChronoUnit.MONTHS);
            }
            else {
                if (this.years != 0) {
                    temporal = temporal.plus((long)this.years, ChronoUnit.YEARS);
                }
                temporal = temporal.plus((long)this.months, ChronoUnit.MONTHS);
            }
        }
        if (this.days != 0) {
            temporal = temporal.plus((long)this.days, ChronoUnit.DAYS);
            return temporal;
        }
        return temporal;
    }
    
    @Override
    public Temporal subtractFrom(Temporal temporal) {
        this.validateChrono(temporal);
        if (this.months == 0) {
            if (this.years != 0) {
                temporal = temporal.minus((long)this.years, ChronoUnit.YEARS);
            }
        }
        else {
            long monthRange = this.monthRange();
            if (monthRange > 0L) {
                temporal = temporal.minus((long)this.years * monthRange + (long)this.months, ChronoUnit.MONTHS);
            }
            else {
                if (this.years != 0) {
                    temporal = temporal.minus((long)this.years, ChronoUnit.YEARS);
                }
                temporal = temporal.minus((long)this.months, ChronoUnit.MONTHS);
            }
        }
        if (this.days != 0) {
            temporal = temporal.minus((long)this.days, ChronoUnit.DAYS);
            return temporal;
        }
        return temporal;
    }
    
    private void validateChrono(TemporalAccessor obj) {
        Objects.requireNonNull(obj, "temporal");
        Chronology chronology = (Chronology)obj.query(TemporalQueries.chronology());
        if (chronology != null && !this.chrono.equals(chronology)) {
            throw new DateTimeException("Chronology mismatch, expected: " + this.chrono.getId() + ", actual: " + chronology.getId());
        }
        return;
    }
    
    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if ((o instanceof ChronoPeriodImpl)) {
            ChronoPeriodImpl chronoPeriodImpl = (ChronoPeriodImpl)o;
            return this.years == chronoPeriodImpl.years && this.months == chronoPeriodImpl.months && this.days == chronoPeriodImpl.days && this.chrono.equals(chronoPeriodImpl.chrono);
        }
        return false;
    }
    
    @Override
    public int hashCode() {
        return this.years + Integer.rotateLeft(this.months, 8) + Integer.rotateLeft(this.days, 16) ^ this.chrono.hashCode();
    }
    
    @Override
    public String toString() {
        if (this.isZero()) {
            return this.getChronology().toString() + " P0D";
        }
        StringBuilder sb = new StringBuilder();
        sb.append(this.getChronology().toString()).append(' ').append('P');
        if (this.years != 0) {
            sb.append(this.years).append('Y');
        }
        if (this.months != 0) {
            sb.append(this.months).append('M');
        }
        if (this.days != 0) {
            sb.append(this.days).append('D');
        }
        return sb.toString();
    }
    
    protected Object writeReplace() {
        return new Ser((byte)9, this);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws ObjectStreamException {
        throw new InvalidObjectException("Deserialization via serialization delegate");
    }
    
    void writeExternal(DataOutput dataOutput) throws IOException {
        dataOutput.writeUTF(this.chrono.getId());
        dataOutput.writeInt(this.years);
        dataOutput.writeInt(this.months);
        dataOutput.writeInt(this.days);
    }
    
    static ChronoPeriodImpl readExternal(DataInput dataInput) throws IOException {
        return new ChronoPeriodImpl(Chronology.of(dataInput.readUTF()), dataInput.readInt(), dataInput.readInt(), dataInput.readInt());
    }
    
    static {
        SUPPORTED_UNITS = Collections.unmodifiableList(Arrays.asList(new TemporalUnit[] { ChronoUnit.YEARS, ChronoUnit.MONTHS, ChronoUnit.DAYS }));
    }
}
